(module cpstrans
   (import js-nodes
	   js-parser
	   js-out
	   js-nodes-extra
	   vardecl
	   config
	   verbose
	   walk)
   (export (cps-transform! tree)))

(define (empty-env)
   (let ((ht (make-hashtable)))
      (hashtable-put! ht 'breaks '())
      (hashtable-put! ht 'continues '())
      (hashtable-put! ht 'return #f)
      (hashtable-put! ht 'label #f)
      ht))

(define (return-env k)
   (let ((ht (make-hashtable)))
      (hashtable-put! ht 'breaks '())
      (hashtable-put! ht 'continues '())
      (hashtable-put! ht 'return k)
      (hashtable-put! ht 'label #f)
      ht))

(define (set-label env label)
   (if (hashtable? env)
       (hashtable-put! env 'label label)
       (error "set-label"
	      "env is NOT a hashtable!" #f)))

(define (take-label env)
   (if (hashtable? env)
       (let ((label (hashtable-get env 'label)))
	  (hashtable-put! env 'label #f)
	  label)
       (error "take-label"
	      "env is NOT a hashtable!" #f)))
   
		     
(define (cps-transform! tree)  ;; 'tree' should be in vars-normal
			       ;; form, although we do not verify it.
   (cps-trans! tree (empty-env)))

(define (push-stack-break env k)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (set! breaks (cons k breaks))
	  (hashtable-put! env 'breaks breaks))))


(define (push-stack-labeled-break env label k)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (set! breaks (cons (cons label k) breaks))
	  (hashtable-put! env 'breaks breaks))))


(define (pop-stack-all-break env)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (set! breaks (cdr breaks))
	      (error "pop-stack-all-break"
		     "No more breaks to pop"
		     #f))
	  (hashtable-put! env 'breaks breaks))))


(define (push-stack-continue env k )
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (set! continues (cons k continues))
	  (hashtable-put! env 'continues continues))))

(define (push-stack-labeled-continue env label k)
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (set! continues (cons (cons label k) continues))
	  (hashtable-put! env 'continues continues))))
   

(define (pop-stack-all-continue env)
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (set! continues (cdr continues))
	      (error "pop-stack-all-continue"
		     "No more continues to pop"
		     #f))
	  (hashtable-put! env 'continues continues))))

(define (search-labeled-break env label)
   (if (hashtable? env)
       (let loop ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (let ((top (car breaks)))
		 (if (and (pair? top)
			  (equal? label (car top)))
		     (cdr top)
		     (loop (cdr breaks))))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (peek-break env)
   (if (hashtable? env)
       (let ((breaks (hashtable-get env 'breaks)))
	  (if (pair? breaks)
	      (let ((top (car breaks)))
		 (if (pair? top)
		     (cdr top)
		     top))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (peek-continue env)
   (if (hashtable? env)
       (let ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (let ((top (car continues)))
		 (if (pair? top)
		     (cdr top)
		     top))
	      #f))
       (error "search-labelled-break" "Not a hashtable!" #f)))

(define (search-labeled-continue env label)
   (if (hashtable? env)
       (let loop ((continues (hashtable-get env 'continues)))
	  (if (pair? continues)
	      (let ((top (car continues)))
		 (if (and (pair? top)
			  (equal? label (car top)))
		     (cdr top)
		     (loop (cdr continues))))
	      #f))
       (error "search-labelled-continue" "Not a hashtable!" #f)))
    
   

(define (get-current-return env)
      (if (hashtable? env)
       (hashtable-get env 'return)
       (error "search-labelled-continue" "Not a hashtable!" #f)))


(define (is-first-var-decl tree) ;; Test if 'tree' is a Block node and
				 ;; the first statement in the node is
				 ;; a Var-Decl-List statement.
   (and (Block? tree)
	(Var-Decl-List? (car (Block-stmts tree)))))
 
(define gen-kvar  ;;Generate a fresh Ref symbol for continuation.
   (let ((n 0))
      (lambda ()
	 (set! n (+ 1 n))
	 (make-Ref (string->symbol (format "_k~a" n))))))

(define gen-var  ;;Generate a fresh Ref symbol for variables.
   (let ((n 0))
      (lambda ()
	 (set! n (+ 1 n))
	 (make-Ref (string->symbol (format "_v~a" n))))))

(define dummy-cont  ;;This is a dummy continuation that does
		      ;;nothing.
   (make-Continu '() (make-Block '())))

(define dummy-var  ;;This is a dummy var that will not be consumed by the cont
   (gen-var))

(define dummy-kvar  ;;This is a dummy var that will not be consumed by the cont
   (gen-kvar))


(define (extend-var var suffix)  ;; Extend Ref symbol of variables
				 ;; with suffix.
   (with-access::Ref var (id)
      (let* ((idstr (symbol->string id))
	     (newid (string->symbol (format (string-append idstr "_~a") suffix))))
	 (make-Ref newid))))

(define (lambda-cps vars exprs) ;; Generate a Cps-Code node, vars is a
				;; list of Params or a param, exprs is
				;; a list of nodes to evaluate or a
				;; single node.
   (let ((params (if (list? vars)
		     vars
		     (list vars)))
	 (stmts (if (list? exprs)
		    exprs
		    (list exprs))))
      (make-Cps-Code params
		     (make-Block stmts))))

(define (lambda-cont vars exprs) ;; Generate a Continu node.
   (let ((params (if (list? vars)
		     vars
		     (list vars)))
	 (stmts (if (list? exprs)
		    exprs
		    (list exprs))))
      (make-Continu params
		    (make-Block stmts))))

(define (lambda-ncont kvar vars exprs) ;; Generate a Named Continu node.
   (let ((params (if (list? vars)
		     vars
		     (list vars)))
	 (stmts (if (list? exprs)
		    exprs
		    (list exprs))))
      (make-Named-Continu kvar
			  (make-Continu params
					(make-Block stmts)))))


(define (apply-cps fun args) ;; Generate a call to Cps-Code node.
   (let ((largs (if (list? args)
		   args
		   (list args))))
      (make-Call-Cps fun largs)))

(define (apply-cont fun args) ; Generate a call to Continu node.
   (let ((largs (if (list? args)
		   args
		   (list args))))
      (make-Call-Cont fun largs)))


;Javascript Library bindings ;; They are considered as Cps-Code.
                             ;; CONSIDER TO MAKE THEM SPECIAL NODE TO
                             ;; PROVIDE INFO FOR OPTIMISATIONS.

(define FunAccess
   (make-Lib-Binding '__ACCESS))

(define FunFuncall
   (make-Lib-Binding '__FUNCALL))

(define FunAccSig
   (make-Lib-Binding '__ACCSIG))

(define FunMethodcall
   (make-Lib-Binding '__METHODCALL))

(define FunNew
   (make-Lib-Binding '__NEW))

(define FunGetref
   (make-Lib-Binding '__GETREF))

(define FunMakefun
   (make-Lib-Binding '__MAKEFUN))

(define LocalThis
   (make-Lib-Binding '_this))

;;========== Node =====================
(define-nmethod (Node.cps-trans!) ;; Generic method for cps-trans! on
				  ;; Node type.
   (error "cps-trans!"
	  "Forget node type:"
	  (class-name (object-class this))))


;;;;;;;;;;;;;;;;;;;;;;Statements;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;========== Program ===================
(define-nmethod (Program.cps-trans!)  ;; Test is the first statement
				      ;; is a variable declaration.
				      ;; Convert the rest to cps
				      ;; style, and apply it to the
				      ;; dummy continuation.
   (with-access::Program this (body)
;      (if (is-first-var-decl body)
;	  (let* ((stmts (Block-stmts body))
;		 (var-decl (car stmts))
;		 (rest-stmts (cdr stmts)))
;	     (set! body
;		   (make-Block (list var-decl
;				     (make-Call
;				      (cps-trans! (make-Block rest-stmts) env)
;				      (list dummy-cont))))))
;	  (set! body
;		(make-Block (list (make-Call
;				      (cps-trans! body env)
;				      (list dummy-cont))))))
;     (set! body (cps-trans! body env))
      (default-walk! this)
      this))

;;========== Block  ===================
(define-nmethod (Block.cps-trans!) ;;Block node.
   (with-access::Block this (stmts)
      (let ((k (gen-kvar)))
	 (lambda-cps k
		     (cps-trans-stmts env stmts k))))) 

(define (cps-trans-stmts env stmts k) ;;Recursively convert stmts list. 
   (let ((v (gen-var)))
      (if (pair? stmts) 
	  (apply-cps (cps-trans! (car stmts) env)
		     (lambda-cont v
				  (cps-trans-stmts env (cdr stmts) k)))
	  (apply-cont k           
		      (make-Undefined 0)))))

;;========== Ref ===================
(define-nmethod (Ref.cps-trans!) ;;Ref node
                                 ;;OTHER INTERPRETATION OF REF NODE
   (let ((k (gen-kvar)))
      (lambda-cps k
		  (apply-cont k
			      this))))


;;========== This ===================
(define-nmethod (This.cps-trans!)
   (cps-trans! LocalThis env))

;;========== Decl ===================
;; c.f. Var-Decl-List

;;========== Var-Decl-List ===================
;;We are considering there normal form here, where Val-Decl-List is
;; dealed in Program and Fun.

;;========== NOP  ===================  ;; 
(define-nmethod (NOP.cps-trans!) ;; NOP node.
   (let ((k (gen-kvar)))
      (lambda-cps k            
		     (apply-cont k
				 (make-Undefined 0)))))

;;========== If  ===================  ;; NOT TESTED
(define-nmethod (If.cps-trans!)
   (with-access::If this (test then else)
      (let ((k (gen-kvar))
	    (v (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! test env)
				(lambda-cont v
					     (make-If v
						      (apply-cps (cps-trans! then env)
								 k)
						      (apply-cps (cps-trans! else env)
								 k))))))))
	   


;;========== For  ===================  ;; NOT TESTED
(define-nmethod (For.cps-trans!)
   (with-access::For this (init test incr body)
      (let ((k (gen-kvar)))
	 (if (not init)
	     (set! init (make-Bool "true")))
	 (if (not test)
	     (set! test (make-Bool "true")))
	 (if (not incr)
	     (set! incr (make-Bool "true")))
	 (lambda-cps k
		     (apply-cps (cps-trans! init env)
				(make-for-cont env test incr body k))))))

(define (make-for-cont env test incr body k)
   (let ((k-test-body (gen-kvar))
	 (k-continue (gen-kvar))
	 (v (gen-var)))
      (let ((label (take-label env)))  ;; push break and continue info
	 (if label
	     (begin
		(push-stack-labeled-break env label k)
		(push-stack-labeled-continue env label k-continue))
	     (begin
		(push-stack-break env k)
		(push-stack-continue env k-continue))))      
      (let ((cps-code (lambda-cont
		       v
		       (list
			(make-Var-Decl-List
			 (list (make-Init
				k-test-body
				(lambda-cont dummy-var
					     (apply-cps (cps-trans! test env)
							(lambda-cont v
								     (make-If v
									      (apply-cps (cps-trans! body env)
											 k-continue)
									      (apply-cont k (make-Undefined 0)))))))
			       (make-Init
				k-continue
				(lambda-cont dummy-var
					     (apply-cps (cps-trans! incr env)
							k-test-body)))))
			(apply-cont k-test-body (make-Undefined 0))))))
	 (pop-stack-all-break env)
	 (pop-stack-all-continue env)
	 cps-code)))

;;========== For-In ===================
(define-nmethod (For-In.cps-trans!) 
   (error "cps-trans!"
	  "For-In statement not supported yet!"
	  (class-name (object-class this))))

;;========== While  ===================
(define-nmethod (While.cps-trans!)
   (with-access::While this (test body)
      (let ((k (gen-kvar))
	    (k-while (gen-kvar))
	    (v (gen-var)))
	 (let ((label (take-label env)))  ;; push break and continue info
	    (if label
		(begin
		   (push-stack-labeled-break env label k)
		   (push-stack-labeled-continue env label k-while))
		(begin
		   (push-stack-break env k)
		   (push-stack-continue env k-while))))
	 (let ((cps-code
		(lambda-cps k
			    (apply-cont (lambda-ncont
					 k-while
					 dummy-var
					 (apply-cps (cps-trans! test env)
						    (lambda-cont v
								 (make-If
								  v
								  (apply-cps (cps-trans! body env)
									     k-while)
								  (apply-cont k (make-Undefined 0))))))
					(make-Undefined 0)))))
	    (pop-stack-all-break env)
	    (pop-stack-all-continue env)
	    cps-code))))



;;========== Do ===================
(define-nmethod (Do.cps-trans!)
   (with-access::Do this (body test)
     (let ((k (gen-kvar))
	    (k-while (gen-kvar))
	    (v (gen-var)))
	 (let ((label (take-label env)))  ;; push break and continue info
	    (if label
		(begin
		   (push-stack-labeled-break env label k)
		   (push-stack-labeled-continue env label k-while))
		(begin
		   (push-stack-break env k)
		   (push-stack-continue env k-while))))
	 (let ((cps-code
		(lambda-cps k
			    (list 
			     (make-Var-Decl-List
			      (list (make-Init k-while
					       (lambda-cont
						dummy-var
						(apply-cps (cps-trans! test env)
							   (lambda-cont v
									(make-If
									 v
									 (apply-cps (cps-trans! body env)
										    k-while)
									 (apply-cont k (make-Undefined 0)))))))))
			     (apply-cps (cps-trans! body env)
					k-while)))))
	    (pop-stack-all-break env)
	    (pop-stack-all-continue env)
	    cps-code))))

;;========== Continue  =================== ;pre-tested
(define-nmethod (Continue.cps-trans!)
   (with-access::Continue this (id)
      (let ((k (gen-kvar))
	    (k-continue (if (not id)
			 (peek-continue env)
			 (search-labeled-continue env id))))
	 (if k-continue
	     (lambda-cps k
			 (apply-cont k-continue (make-Undefined 0)))
	     (error "cps-trans!"
		    "Invalid continue statement!"
		    (if id id #f))))))

;;========== Break  =================== pre-tested
(define-nmethod (Break.cps-trans!)
   (with-access::Break this (id)
      (let ((k (gen-kvar))
	    (k-break (if (not id)
			 (peek-break env)
			 (search-labeled-break env id))))
	 (if k-break
	     (lambda-cps k
			 (apply-cont k-break (make-Undefined 0)))
	     (error "cps-trans!"
		    "Invalid break statement!"
		    (if id id #f))))))
   
;;========== Return  =================== UNFINISHED
(define-nmethod (Return.cps-trans!)
   (with-access::Return this (val)
      (lambda-cps dummy-kvar
		  (apply-cps (cps-trans! val env)
			     (if (get-current-return env)
				 (get-current-return env)
				 (error "cps-trans!"
					"no return point"
					#f))))))
		      
;;========== With ===================
(define-nmethod (With.cps-trans!) 
   (error "cps-trans!"
	  "With statement not supported yet!"
	  (class-name (object-class this))))

;;========== Switch =================== UNFINISHED
(define-nmethod (Switch.cps-trans!)
   (with-access::Switch this (key cases)
      (let ((k (gen-kvar)))
	 (push-stack-break env k)
	 (let ((cps-code (lambda-cps k
				     (apply-cps (cps-trans! key env)
						(cps-trans-switch-body env cases k)))))
	    (pop-stack-all-break env)
	    cps-code))))


(define (cps-trans-switch-body env body k)
   (let ((v (gen-var))
	 (case-var (gen-var))
	 (n 0)
	 (default? #f))
      (lambda-cont v
		   (if (pair? body)
		       (list
			(make-Var-Decl-List
			 (map (lambda (a_case)
				 (let ((rel (make-Init (extend-var case-var n)
						       (lambda-cont dummy-var
								   (apply-cps 
								    (if (Case? a_case)
									(cps-trans! (Case-body a_case) env)
									(cps-trans! (Default-body a_case) env))
								    (extend-var case-var (+ n 1)))))))
				    
				    (set! n (+ n 1))
				    ;(if (Default? a_case)
				    ;	(begin
				    ;	   (set! default? #t)
				    ;	   (tprint "found default")))
				    rel))
			      body))
			(make-Var-Decl-List (list (make-Init (extend-var case-var n) k)))
			(cps-trans-switch-cases env body 0 k v case-var #f))
		       (apply-cont k v)))))

(define (cps-trans-switch-cases env body dep k v case-var default)
   (let ((v1 (gen-var)))
      (if (pair? body)
	  (if (Case? (car body))
	      (apply-cps
	       (cps-trans! (Case-expr (car body)) env)
	       (lambda-cont v1
			    (make-If (make-Binary v '=== v1)
				     (apply-cont
				      (extend-var case-var dep) (make-Undefined 0))
				     (cps-trans-switch-cases env (cdr body) (+ dep 1) k v case-var default))))
	      (cps-trans-switch-cases env (cdr body) (+ dep 1) k v case-var dep))
	  (if default
	      (apply-cont
	       (extend-var case-var default) (make-Undefined 0))
	      (apply-cont
	       k (make-Undefined 0))))))
	      
				
       

;;========== Case  =================== 
;; c.f. case of Switch

;;========== Default =================== 
;; c.f. case of Default

;;========== Throw =================== unsupported yet

;;========== Try  =================== unsupported yet

;;========== Catch =================== unsupported yet

;;========== Labeled =================== ;pre-tested
(define-nmethod (Labeled.cps-trans!)
   (with-access::Labeled this (id body)
      (if (or (For? body)
	      (While? body)
	      (Do? body))
	  (begin
	     (set-label env id)
	     (cps-trans! body env))
	  (cps-trans! body env))))
;;========== Fun-Binding =================== 
;; Not possible in the normal form

;;========== Assig  ===================
;; C.f. sub-cases

;;========== Assig-op  =================== 
;; C.f. sub-cases

;;========== Vassig  ===================
(define-nmethod (Vassig.cps-trans!) ;;Vassig node 
   (with-access::Vassig this (lhs rhs)
      (let* ((k (gen-kvar))
	     (v (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! rhs env)
				(lambda-cont v
					     (list
					      (make-Vassig lhs v)
					      (apply-cont k v))))))))

;;========== Accsig  ===================
(define-nmethod (Accsig.cps-trans!) ;;Accsig node 
   (with-access::Accsig this (lhs rhs)
      (with-access::Access lhs (obj field)
	 (let ((k (gen-kvar))
	       (v1 (gen-var))
	       (v2 (gen-var))
	       (v3 (gen-var)))
	    (lambda-cps
	     k
	     (apply-cps
	      (cps-trans! obj env)
	      (lambda-cont
	       v1
	       (apply-cps
		(cps-trans! field env)
		(lambda-cont
		 v2
		 (apply-cps
		  (cps-trans! rhs env)
		  (lambda-cont
		   v3
		   ;(apply-cont k (make-Accsig (make-Access v1 v2) v3))
		   (apply-cps FunAccSig (list k v1 v2 v3))
		   )))))))))))



;;========== Vassig-op  =================== pre-tested
(define-nmethod (Vassig-op.cps-trans!) ;;Vassig node 
   (with-access::Vassig-op this (lhs rhs op)
      (let* ((k (gen-kvar))
	     (v (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! rhs env)
				(lambda-cont v
					     (list
					      (make-Vassig-op lhs v op)
					      (apply-cont k v))))))))

;;========== Accsig-op  =================== pre-tested
(define-nmethod (Accsig-op.cps-trans!) ;;Accsig node 
   (with-access::Accsig-op this (lhs rhs op)
      (with-access::Access lhs (obj field)
	 (let ((k (gen-kvar))
	       (v1 (gen-var))
	       (v2 (gen-var))
	       (v3 (gen-var)))
	    (lambda-cps
	     k
	     (apply-cps
	      (cps-trans! obj env)
	      (lambda-cont
	       v1
	       (apply-cps
		(cps-trans! field env)
		(lambda-cont
		 v2
		 (apply-cps
		  (cps-trans! rhs env)
		  (lambda-cont
		   v3
		   (apply-cont
		    k
		    (make-Accsig-op (make-Access v1 v2) v3 op)))))))))))))



;;========== Inits  ===================
;; C.f. Var-Decl-List

;;========== Param  ===================
;; C.f. other cases

;;========== Named-Fun =================== Pre-rested
(define-nmethod (Named-Fun.cps-trans!)
   (with-access::Named-Fun this (name fun)
      (cps-trans! fun env)))


;;========== Fun  ===================


(define-nmethod (Var-Fun.cps-trans!) ;; haven't tested yet. Similar to the
				 ;; one on Program node. CHEKC HOW TO
				 ;; DEAL WITH RETURN STATEMENT
   (with-access::Var-Fun this (params body var-decl)
      (let* ((k1 (gen-kvar))
	     (k (gen-kvar))
	     (v (gen-var))
	     (new-env (return-env k)))
	 (lambda-cps k1
		     (apply-cont
		      k1
		      (make-Call FunMakefun
				 (list (make-Var-Fun (cons k params)
						     (apply-cps (cps-trans! body new-env) k)
						     var-decl))))))))


;;========== Sequence  =================== ;;Pre-tested
(define-nmethod (Sequence.cps-trans!)
   (with-access::Sequence this (exprs)
      (let ((k (gen-kvar)))
	 (lambda-cps k
		     (cps-trans-exprs env exprs k)))))

(define (cps-trans-exprs env exprs k) ;;Recursively convert stmts list. 
   (let ((v (gen-var)))
      (cond ((pair? (cdr exprs)) 
	     (apply-cps (cps-trans! (car exprs) env)
			(lambda-cont v
				     (cps-trans-exprs env (cdr exprs) k))))
	    ((and (pair? exprs) (not (pair? (cdr exprs))))
	     (apply-cps (cps-trans! (car exprs) env)
			(lambda-cont v
				     (apply-cont k v))))
	    (else
	     (error "cps-trans!"
		    "Missing case in cps-trans-exprs."
		    #f)))))

;;========== Cond  =================== ;; Pre-tested
(define-nmethod (Cond.cps-trans!)
   (with-access::Cond this (test then else)
      (let ((k (gen-kvar))
	    (v (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! test env)
				(lambda-cont v
					     (make-If v
						      (apply-cps (cps-trans! then env)
								 k)
						      (apply-cps (cps-trans! else env)
								 k))))))))


;;========== Binary  ===================
(define-nmethod (Binary.cps-trans!)
   (with-access::Binary this (lhs op rhs)
      (let ((k (gen-kvar))
	    (vl (gen-var))
	    (vr (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! lhs env)
				(lambda-cont vl
					     (apply-cps (cps-trans! rhs env)
							(lambda-cont vr
								     (apply-cont k
										 (make-Binary vl op vr))))))))))

;;========== Unary  ===================
(define-nmethod (Unary.cps-trans!) ;; Unary node
   (with-access::Unary this (op expr)
      (let ((k (gen-kvar))
	    (v (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! expr env)
				(lambda-cont v
					     (apply-cont k
							 (make-Unary op v))))))))

;;========== Postfix =================== ;;IFRAME OBJECT NOT SUPPORTED, AS ASSIGNMENT
(define-nmethod (Postfix.cps-trans!)
   (with-access::Postfix this (expr op)
      (cond ((Access? expr)
	     (with-access::Access expr (obj field)
		(let ((k (gen-kvar))
		      (v1 (gen-var))
		      (v2 (gen-var)))
		   (lambda-cps k
			       (apply-cps (cps-trans! obj env)
					  (lambda-cont v1
						       (apply-cps (cps-trans! field env)
								  (lambda-cont v2
									       (apply-cont k
											   (make-Postfix (make-Access v1 v2) op))))))))))
	    ((Ref? expr)
	     (let ((k (gen-kvar)))
		   (lambda-cps k
			       (apply-cont k this))))
	    (else
	     (error "cps-trans!"
		    "Missing case in Postfix.cps-trans!"
		    #f)))))

;;========== New  ===================

(define-nmethod (New.cps-trans!) ;;New node
   (with-access::New this (class args)
      (let ((k (gen-kvar))
	    (v (gen-var)))
	 (lambda-cps k
	     (apply-cps (cps-trans! class env)
			(cps-trans-new env args v 0 k))))))



(define (cps-trans-new env args v dep k) ; args is the list of arguments to new 
   (if (pair? args) ;if not null;
       (lambda-cont (extend-var v dep)
		    (apply-cps
		     (cps-trans! (car args) env)
		     (cps-trans-new env (cdr args) v (+ dep 1) k)))
       (lambda-cont (extend-var v dep)
		    (apply-cps FunNew
			       (cons k
				     (map (lambda (n) (extend-var v n))
					  (iota (+ dep 1) 0 1)))))))


;;========== Access  ===================
(define-nmethod (Access.cps-trans!) ;;Access Node
   (with-access::Access this (obj field)
      (let ((k (gen-kvar))
	    (v1 (gen-var))
	    (v2 (gen-var)))
	 (lambda-cps k
		     (apply-cps (cps-trans! obj env)
				(lambda-cont v1
					     (apply-cps (cps-trans! field env)
							(lambda-cont v2
								     (apply-cps FunAccess
										(list k v1 v2))))))))))

;;========== Call  ===================
(define-nmethod (Call.cps-trans!) ;;Call Node. Before calling the
				  ;;function or the method, check if
				  ;;it is a cps-enabled function or
				  ;;not.
   (with-access::Call this (fun args)
      (let* ((k (gen-kvar))
	     (v (gen-var)))
	 (if (Access? fun)
	     (lambda-cps k
			 (apply-cps (cps-trans! (Access-obj fun) env)
				    (lambda-cont (extend-var v 0)
						 (apply-cps (cps-trans! (Access-field fun) env)
							    (cps-trans-method-call env args v 1 k)))))
	     (lambda-cps k
			 (apply-cps (cps-trans! fun env)
				    (cps-trans-fun-call env args v 0 k)))))))


(define (cps-trans-method-call env args v dep k)
   (if (pair? args)
       (lambda-cont (extend-var v dep)
		    (apply-cps (cps-trans! (car args) env)
			       (cps-trans-method-call env (cdr args) v (+ dep 1) k)))
       (lambda-cont (extend-var v dep)
		    (apply-cps FunMethodcall
			       (cons k
				     (map (lambda (n) (extend-var v n))
					  (iota (+ dep 1) 0 1)))))))

(define (cps-trans-fun-call env args v dep k)
   (if (pair? args)
       (lambda-cont (extend-var v dep)
		    (apply-cps (cps-trans! (car args) env)
			       (cps-trans-fun-call env (cdr args) v (+ dep 1) k)))
       (lambda-cont (extend-var v dep)
		    (apply-cps FunFuncall
			       (cons k
				     (map (lambda (n) (extend-var v n))
					  (iota (+ dep 1) 0 1)))))))

;;========== Pragma  ===================
(define-nmethod (Pragma.cps-trans!) 
   (error "cps-trans!"
	  "Finally I can see what is a Pragma node in Javascript"
	  (class-name (object-class this))))

;;========== Array ===================
(define-nmethod (Array.cps-trans!) ;;Array node
   (with-access::Array this (els len)
      (let* ((k (gen-kvar))
	     (v (gen-var)))
	 (lambda-cps k
		     (cps-trans-array env els v 0 k)))))

(define (cps-trans-array env els v dep k)   ;; CHECK IF THE CASE OF EMPTY ELEMENTS ARE COVERED 
   (if (pair? els)
       (apply-cps (cps-trans! (Array-Element-expr (car els)) env)
		  (lambda-cont (extend-var v dep)
			       (cps-trans-array env (cdr els) v (+ dep 1) k)))
       (apply-cont k
		   (make-Array (map (lambda (n) (make-Array-Element n (extend-var v n)))
				    (iota dep 0 1))
			       dep))))

;;========== Array-Element ===================
;; C.f. the case of Array.

;;========== Obj-Init  ===================

(define-nmethod (Obj-Init.cps-trans!) ;; Obj-Init
   (with-access::Obj-Init this (inits) ;inits is a list of Property-Init
      (let* ((k (gen-kvar))
	     (v (gen-var)))
	 (lambda-cps k
		     (cps-trans-inits env inits v 0 k this)))))


(define (cps-trans-inits env inits v dep k obj-inits) ; inits is the list of Property-Init 
   (if (pair? inits)
       (with-access::Property-Init (car inits) (name val)
	  (apply-cps (cps-trans! val env)
		     (lambda-cont (extend-var v dep)
				  (cps-trans-inits env (cdr inits) v (+ dep 1) k obj-inits))))
       (apply-cont k
		   (let ((n -1))
		      (make-Obj-Init               ;; Rewrite this mapping to a filtering
		       (map (lambda (node)
			       (with-access::Property-Init node (name val)
				  (set! n (+ n 1))
				  (make-Property-Init name (extend-var v n))))
			    (Obj-Init-inits obj-inits)))))))

;;========== Property-Init  ===================
;; C.f. the case of Obj-Init.

;;========== Literal  ===================
(define-nmethod (Literal.cps-trans!) ;; Literal
   (let ((k (gen-kvar)))
      (lambda-cps k
		  (apply-cont k this))))




