(module jscps
   (import js-nodes
	   js-parser
	   js-out
	   vardecl
	   config
	   verbose
	   walk)
   (export (cps-transform js-prog-ast)
	   (bootstrap cps-out varlist)
	   (warpcode funname code-ast)))

;Javascript Library bindings
(define FunAccess
   (make-Ref '__ACCESS))

(define FunFuncall
   (make-Ref '__FUNCALL))

(define FunMethodcall
   (make-Ref '__METHODCALL))

(define FunNew
   (make-Ref '__NEW))

(define FunGetref
   (make-Ref '__GETREF))

(define FunMakefun
   (make-Ref '__MAKEFUN))

(define (cps-transform js-prog-ast)
   (if (not (Program? js-prog-ast))
       (error "cps-transform"
	      "Not a program!"
	      ""))
   (let* ((body (Program-body js-prog-ast))
	  (normal-form body)
	  (stmts (Block-stmts normal-form)))
      (if (Var-Decl-List? (car stmts))
	  (let ((var-decl (car stmts))
		(cps-result (cps-pass (make-Block (cdr stmts)))))
	     (make-Program
	      (make-Block
	       (list var-decl                                 ;; initialize global variables
		     (make-Var-Decl-List                      ;; initialize __cpscode
		      (list (make-Init (make-Ref '__cpscode)
				       cps-result)))))))
	  (let ((cps-result (cps-pass body)))
	     (make-Program
	      (make-Block
	       (list (make-Var-Decl-List                      ;;initialize __cpscode
		      (list (make-Init (make-Ref '__cpscode)
				       cps-result))))))))))
       
(define (warp-string str)                                     ;; CAUTION: make a better quote here!!
   (make-String (format "'~a'" str)))

(define (warpcode funname code-ast)
   (make-Program
    (make-Block
     (list
      (make-Fun-Binding (make-Decl funname)
			(make-Fun '()
				  (Program-body code-ast)))))))

(define (bootstrap-aux varlist kvar)
   (let ((vvar (genvart)))
      (if (pair? varlist)
	  (cps-apply FunGetref                      ;;Call __GETREF
		     (list
		      (warp-string (cdar varlist))  ;; with vart
		      (cps-lambda                   ;; and this continuation 
		       vvar                        
		       (make-Block           
			(list
			 (make-Vassig (make-Ref (caar varlist)) vvar) ;;assign result to varl
			 (bootstrap-aux (cdr varlist) kvar))))))           ;;continute to bootstrap
	  (cps-apply (make-Ref '__cpscode)                              ;;no more bootstraping, invoke cpscode
		     kvar))))              ;;a dummy continuation

(define (bootstrap cps-prog-ast varlist)
   (let ((kvar (genkont)))
      (if (pair? varlist)
	  (make-Program
	   (make-Block       ;;make statements
	    (append
	     (if (pair? varlist)
		 (list
		  (make-Var-Decl-List                   ;;declare local variables for foreign object
		   (map (lambda (p)
			   (make-Decl (car p)))
			varlist)))
		 '())
	     (Block-stmts (Program-body cps-prog-ast)) 
	     (list (make-Var-Decl-List
		    (list (make-Init (make-Ref '__bootstrap)
				     (make-Fun (list kvar)
					       (make-Block (list (bootstrap-aux varlist kvar)))))))))))
	  cps-prog-ast)))
     
(define genkont
   (let ((n 0))
      (lambda ()
	 (set! n (+ 1 n))
	 (make-Ref (string->symbol (format "_k~a" n))))))

(define genvart
   (let ((n 0))
      (lambda ()
	 (set! n (+ 1 n))
	 (make-Ref (string->symbol (format "_v~a" n))))))

(define (extendvart var suffix)
   (with-access::Ref var (id)
      (let* ((idstr (symbol->string id))
	     (newid (string->symbol (format (string-append idstr "_~a") suffix))))
	 (make-Ref newid))))

(define-generic (cps-pass node::Node)
   (tprint node)
   (error "cps-pass"
	  "Unsupported node type"
	  (class-name (object-class node))))

;(define-method (cps-pass node::Program)
;   (with-access::Program node (body)
;      (let ((cps-result (cps-pass body)))
;	 (set! body cps-result)
;	 node)))

;(define-method (cps-pass node::Block)
;   (with-access::Block node (stmts)
;      (let ((cps-result (cps-pass-stmts stmts)))
;	 (set! stmts (list cps-result))
;	 node)))
;
;(define-method (cps-pass node::Program)
;   (with-access::Program node (body)
;      (let ((cps-result (cps-pass body)))
;	 cps-result)))

(define-method (cps-pass node::Block)
   (with-access::Block node (stmts)
      (let ((cps-result (cps-pass-stmts stmts)))
	 cps-result)))

(define (cps-lambda lvars body) ;lvars is possibly a list of symbols
   (let ((params (if (list? lvars)
		     lvars
		     (list lvars))))
	  (make-Fun params
		    (make-Block (list body)))))

(define (cps-apply fun largs) ; largs is possibly a list of expressions
   (let ((args (if (list? largs)
		   largs
		   (list largs))))
      (make-Call fun args)))

(define (cps-pass-stmts stmts)
   (let* ((kvar (genkont))
	 (vvar (genvart)))
      (if (pair? (cdr stmts)) 
	  (cps-lambda            ; If it is not the last statement in the list
	     kvar
	     (cps-apply
	        (cps-pass (car stmts))
		(cps-lambda
		   vvar
		   (cps-apply
		      (cps-pass-stmts (cdr stmts))
		      kvar))))
	  (cps-lambda           ; If it is the last statement in the list
	     kvar
	     (cps-apply
	        (cps-pass (car stmts))
		kvar)))))



(define-method (cps-pass node::Ref)
   (with-access::Ref node (id)
      (let* ((kvar (genkont)))
	 (cps-lambda
	    kvar
	    (cps-apply
	       kvar
	       node)))))

(define-method (cps-pass node::Vassig)
   (with-access::Vassig node (lhs rhs)
   (let* ((kvar (genkont))
	 (vvar (genvart)))
      (cps-lambda
          kvar
	  (cps-apply
	      (cps-pass rhs)
	      (cps-lambda
	          vvar
		  (make-Block
		   (list
		    (make-Vassig lhs vvar)
		    (cps-apply kvar vvar)))))))))


	       
	       

(define (cps-pass-new largs v depth k) ; largs is the list of arguments to new 
   (if (pair? largs) ;if not null;
       (cps-lambda
	  (extendvart v depth)
	  (cps-apply
	      (cps-pass (car largs))
	      (cps-pass-new (cdr largs) v (+ depth 1) k)))
       (cps-lambda
	  (extendvart v depth)
	  (cps-apply
	     FunNew
	     (cons k
		   (map (lambda (n) (extendvart v n))
			(iota (+ depth 1) 0 1)))))))

(define-method (cps-pass node::New)
   (with-access::New node (class args)
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda
	     kvar
	     (cps-apply
	         (cps-pass class)
		 (cps-pass-new args vvar 0 kvar))))))

(define-method (cps-pass node::Access)
   (with-access::Access node (obj field)
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda
	    kvar
	    (cps-apply
	       (cps-pass obj)
	       (cps-lambda
		  vvar
		  (cps-apply FunAccess
			     (list kvar vvar field))))))))

(define-method (cps-pass node::Literal)
   (with-access::Literal node (val)
      (let* ((kvar (genkont)))
	 (cps-lambda
	    kvar
	    (cps-apply
	       kvar
	       node)))))

(define (cps-pass-inits inits v depth k root) ; inits is the list of Property-Init 
   (if (pair? inits)
       (with-access::Property-Init (car inits) (name val)
	  (cps-apply (cps-pass val)
		     (cps-lambda (extendvart v depth)
				 (cps-pass-inits (cdr inits) v (+ depth 1) k root))))
       (cps-apply k
		  (let ((n -1))
		     (make-Obj-Init
		      (map (lambda (node)
			      (with-access::Property-Init node (name val)
				 (set! n (+ n 1))
				 (make-Property-Init name (extendvart v n))))
			   (Obj-Init-inits root)))))))
			  

(define-method (cps-pass node::Obj-Init)
   (with-access::Obj-Init node (inits) ;inits is a list of Property-Init
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda
	    kvar
	    (cps-pass-inits inits vvar 0 kvar node)))))
      
(define (cps-pass-method-call largs v depth k)
   (if (pair? largs)
       (cps-lambda (extendvart v depth)
		   (cps-apply (cps-pass (car largs))
			      (cps-pass-method-call (cdr largs) v (+ depth 1) k)))
       (cps-lambda (extendvart v depth)
		   (cps-apply FunMethodcall
			      (cons k
				    (map (lambda (n) (extendvart v n))
					 (iota (+ depth 1) 0 1)))))))

(define (cps-pass-fun-call largs v depth k)
   (if (pair? largs)
       (cps-lambda (extendvart v depth)
		   (cps-apply (cps-pass (car largs))
			      (cps-pass-fun-call (cdr largs) v (+ depth 1) k)))
       (cps-lambda (extendvart v depth)
		   (cps-apply FunFuncall
			      (cons k
				    (map (lambda (n) (extendvart v n))
					 (iota (+ depth 1) 0 1)))))))

   
(define-method (cps-pass node::Call)
   (with-access::Call node (fun args)
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (if (Access? fun)
	     (cps-lambda
	        kvar
		(cps-apply
		   (cps-pass (Access-obj fun))
		   (cps-lambda (extendvart vvar 0)
			       (cps-apply
				  (cps-pass (Access-field fun))
				  (cps-pass-method-call args vvar 1 kvar)))))
	     (cps-lambda
	        kvar
		(cps-apply
		   (cps-pass fun)
		   (cps-pass-fun-call args vvar 0 kvar)))))))
				  
	  
(define-method (cps-pass node::Unary)
   (with-access::Unary node (op expr)
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda kvar
		     (cps-apply (cps-pass expr)
				(cps-lambda vvar
					    (cps-apply kvar
						       (make-Unary op vvar))))))))

(define-method (cps-pass node::Fun)
   (with-access::Fun node (params body)
      (let* ((kvar1 (genkont))
	     (kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda kvar1
		     (cps-apply
		      kvar1
		      (cps-apply FunMakefun
				 (make-Fun
				  (cons kvar params)
				  (with-access::Block body (stmts)
				     (if (Var-Decl-List? (car stmts))
					 (let ((var-decl (car stmts))
					       (cps-result (cps-pass (make-Block (cdr stmts)))))
					    (make-Block (list var-decl
							      (cps-apply
							       cps-result
							       (cps-lambda vvar
									   (cps-apply kvar vvar))))))
					 (make-Block (list (cps-apply
							    (cps-pass body)
							    (cps-lambda vvar
									(cps-apply kvar vvar))))))))))))))



(define (cps-pass-array els vvar depth kvar)
   (if (pair? els)
       (cps-apply
	(cps-pass (Array-Element-expr (car els)))
	(cps-lambda (extendvart vvar depth)
		    (cps-pass-array (cdr els) vvar (+ depth 1) kvar)))
       (cps-apply kvar
		  (make-Array (map (lambda (n) (make-Array-Element n (extendvart vvar n)))
			       (iota depth 0 1))
			 depth))))
       

(define-method (cps-pass node::Array)
   (with-access::Array node (els len)
      (let* ((kvar (genkont))
	     (vvar (genvart)))
	 (cps-lambda
	     kvar
	     (cps-pass-array els vvar 0 kvar)))))

