(module js2post-main
   (import config
	   js2post)
   (main my-main))

(define *in-file* #f)
(define *out-file* #f)
(define *var-list* '()); a list of foreign names imported by script (e.g. the google object)


(define *version* "20110210")

(define (handle-args args)
   (args-parse (cdr args)
      (section "Help")
      (("?")
       (args-parse-usage #f)
       (exit))
      ((("-h" "--help") (help "?,-h,--help" "This help message"))
       (args-parse-usage #f)
       (exit))
      (section "Misc")
      ((("--version") (help "Version number"))
       (print *version*)
       (exit))
      ((("-v" "--verbose") (help "Verbose output"))
       (config-set! 'verbose #t))
      (("-o" ?file (help "The output file. '-' prints to stdout."))
       (set! *out-file* file))
      (("-i" ?file (help "The input file. '-' reads from stdin."))
       (set! *in-file* file))
      (("-a" ?varl ?vart (help "'VARL' should point to 'VART' imported from inner iframe"))
       (set! *var-list* (cons (cons (string->symbol varl) vart) *var-list*)))
      (else
       (begin
	  (display "Unknow argument: " (current-error-port))
	  (display else (current-error-port))
	  (args-parse-usage #f)
	  (exit)))))

(define (my-main args)
   (config-init!)
   (handle-args args)
   (let ((in-p (if (not *in-file*)
		   (current-input-port)
		   (open-input-file *in-file*)))
	 (out-p (if (not *out-file*)
		    (current-output-port)
		    (open-output-file *out-file*))))
      (with-handler
	 (lambda (e)
	    (unless (not *out-file*)
	       (delete-file *out-file*))
	    (raise e))
	 (unwind-protect
	    (js2post-compile-file in-p out-p *var-list*)
	    (begin
	       (close-input-port in-p)
 	       (close-output-port out-p))))
      0)
   )
