//message_handler handles incomming messages
function message_handler(e){
    var recv_msg = JSON.parse(e.data);
    var origin = e.origin;
    
    //verbose("Revc from ", origin, " : ", e.data);
    
    var resp_msg = dispatch(recv_msg,e.source,e.origin);
    var resp_json =  JSON.stringify(resp_msg);
    e.source.postMessage(resp_json,origin);
    
    verbose("I=>O Reply"/*, origin*/, " : ", resp_json);
}

//resigister a callback message_handler on 'message' event 
(function (){
    window.addEventListener("message", 
			    message_handler, false);
})();

//A simple list to store mapping from ohandle to 
var add_handle_obj, get_obj_by_handle;
(function(){

    var handle_list = new Object();
    
    add_handle_obj = function(ohandle,obj){
	handle_list[ohandle.__id] = obj;
    };
    
    get_obj_by_handle = function(ohandle){
	return handle_list[ohandle.__id];
    };
})();

//make response message of the form {msg_id: - , msg_type: EXE_CONT | CALLBACK, return_val : -}  
function make_resp_msg(recv,obj){

    var ohandle, msg;

    if (obj != null && (typeof(obj) == 'object'|| typeof(obj) == 'function')){
	ohandle = new OHandle();
	add_handle_obj(ohandle,obj);
	msg = {msg_id : recv.msg_id, msg_type:'EXE_CONT', return_val : ohandle};
    }
    else{
	msg = {msg_id: recv.msg_id, msg_type:'EXE_CONT', return_val : obj};
    }

    return msg;
}

//response of message: {msg_id: -, msg_type: GET_GLOBAL_REF, global_name : -}
function get_global_ref(recv){
    var obj = window[recv.global_name];

    return make_resp_msg(recv,obj);
}

//response of message: {msg_id: -, msg_type: GET_PROPERTY, object : -, property_name : -}
function get_property(recv){
    var obj = get_obj_by_handle(recv.object);
    var property = obj[recv.property_name];
    
    return make_resp_msg(recv,property);
}

function dehandle(obj,source,origin){
    var val;
    if (obj instanceof Array)
	val = new Array();
    else
	val = new Object();

    for( var prop in obj){
	if (obj[prop] != null && obj[prop].__is_fhandle){
	    val[prop] = warp_proxy_fun(obj[prop],source,origin);
	}
	else if(obj[prop] != null && obj[prop].__is_ohandle){
	    val[prop] = get_obj_by_handle(obj[prop]);
	}
	else if(obj[prop] != null && typeof(obj[prop]) == 'object'){
	    val[prop] = dehandle(obj[prop],source,origin);
	}
	else{
	    val[prop] = obj[prop];
	}
    }
    return val;
};

function reconstruct_args(__arguments,source,origin){
    var args = new Array();
    for (var i= 0; i<__arguments.length; i++){
	if ( __arguments[i] != null && __arguments[i].__is_fhandle){
	    args[i] = warp_proxy_fun(__arguments[i],source,origin);
	}
	else if(__arguments[i] != null &&  __arguments[i].__is_ohandle){
	    args[i] = get_obj_by_handle(__arguments[i]);
	}
	else if(__arguments[i] != null && typeof(__arguments[i]) == 'object'){
	    args[i] = dehandle(__arguments[i],source,dehandle);
	}
	else{
	    args[i] = __arguments[i];
	}
    }    
    return args;
}


//response of message: {msg_id: -, msg_type: CALL_METHOD, object : - , method_name : -,  arguments : -}
function call_method(recv,source,origin){
    var obj = get_obj_by_handle(recv.object);
    var method = obj[recv.method_name];
    var __arguments = recv.arguments;
    
    //reconstruct arguments to the method invocation
    var args = reconstruct_args(__arguments,source,origin);
    
    //actually apply the method
    val = method.apply(obj,args);
    
    return make_resp_msg(recv,val);
}

//response of message: {msg_id: -, msg_type: CALL_FUNCTION, function : -,  arguments : -}
function call_function(recv,source,origin){
    var fun = get_obj_by_handle(recv._function);
    var __arguments = recv.arguments;
    
    //reconstruct arguments to the method invocation
    var args = reconstruct_args(__arguments,origin);
    
    //actually apply the method
    val = method.apply(null,args);

    return make_resp_msg(recv,val);
}

//response of message: {msg_id: -, msg_type: NEW_OBJECT, object : - , method : -,  arguments : -}
function new_object(recv,source,origin){
    var fun = get_obj_by_handle(recv._function);
    var __arguments = recv.arguments;
    
    //reconstruct arguments to the method invocation
    var args = reconstruct_args(__arguments,origin);
    var val;

    if (args.length == 0)
	val = new fun();
    else{
	var expr = new String;
	expr = "new fun(";
	for (var i = 0; i<args.length - 1; i++)
	    expr = expr + "args[" + i + "],";
	expr  = expr + "args[" + i + "])";
	val = eval(expr);
    }
    return make_resp_msg(recv,val);
}

function obj_prop_assign(recv,source,origin){
    var obj = get_obj_by_handle(recv.object);
    var prop = recv.property_name;
    var value = recv.value;
    
    //reconstruct
    var new_value = (reconstruct_args([value],source,origin))[0];
    
    //assignment
    obj[prop]=new_value;
    
    return make_resp_msg(recv,true);
}


//dispatch function handles the messages
function dispatch(recv_msg,source,origin){
    switch(recv_msg.msg_type){
    case 'GET_GLOBAL_REF':
	return get_global_ref(recv_msg);
	break;
    case 'GET_PROPERTY':
	return get_property(recv_msg);
	break;
    case 'CALL_METHOD':
	return call_method(recv_msg,source,origin);
	break;
    case 'CALL_FUNCTION':
	return call_function(recv_msg,source,origin);
	break;
    case 'NEW_OBJECT':
	return new_object(recv_msg,source,origin);
	break;
    case 'OBJ_PROP_ASSIGN':
	return obj_prop_assign(recv_msg,source,origin);
    }
}

//warp a proxy callback function
function warp_proxy_fun(fhandle,source,origin){
    var proxy_fun;
    proxy_fun = function(){
	var args = new Array();
	for (var i= 0; i<arguments.length; i++){
	    if (typeof arguments[i] == 'function' || typeof arguments[i] == 'object' ){
		var ohandle = new OHandle();
		add_handle_obj(ohandle,arguments[i]);
		args[i] = ohandle;
	    }
	    else{
		args[i] = arguments[i];
	    }
	}
	var call_msg = {_function : fhandle, msg_type: 'CALLBACK', arguments : args};
	var call_json = JSON.stringify(call_msg);
	verbose("I=>O Callback"/* + origin + */," : " + call_json );
	source.postMessage(call_json,origin);
	return;
    };
    return proxy_fun;
}