(module trampoline
   (import js-nodes
	   js-parser
	   js-out
	   js-nodes-extra
	   config
	   verbose
	   walk)
   (export (trampoline! tree)))

(define make-trampoline-test make-trampoline-test-inlined) ;; default inlined

(define (trampoline! tree)
   (trans-trampoline! tree #f))

(define-nmethod (Node.trans-trampoline!)
   (default-walk! this)
   this)

(define trampoline-counter
   (make-Ref '_c_g))

(define trampoline-limit
   (make-Number 40))

(define trampoline-class
   (make-Ref '__Tramp))

(define trampoline-var
   (make-Ref '_res_or_tramp))

(define gen-tvar  ;;Generate a fresh Ref symbol for variables.
   (let ((n 0))
      (lambda ()
	 (set! n (+ 1 n))
	 (make-Ref (string->symbol (format "_t~a" n))))))

(define (new-init var expr)
   (make-Var-Decl-List
    (list (make-Init var expr))))

(define (simple-expr? expr)
   (or (Ref? expr)
       (Literal? expr)))


(define (make-trampoline-test-inlined expr)
   (if (= (length (Call-args expr)) 1)
       (let* ((fun-var (gen-tvar))
	      (arg-var (gen-tvar))
	      (fun-expr (Call-fun expr))
	      (arg-expr (let ((args (Call-args expr)))
			   (if (> (length args) 1)
			       (error "make-trampoline-test-inlined"
				      "Why there are more than one parameters?"
				      args))
			   (car args)))
	      (trampoline-code
	       (make-Block
		(filter
		 (lambda (x) x)
		 (list
		  (if (not (simple-expr? fun-expr))
		      (new-init fun-var fun-expr)
		      #f)
		  (if (not (simple-expr? arg-expr))
		      (new-init arg-var arg-expr)
		      #f)
		  (let ((fun-final (if (not (simple-expr? fun-expr)) fun-var fun-expr))
			(arg-final (if (not (simple-expr? arg-expr))
				       arg-var
				       arg-expr)))
		     (make-Return
		      (make-Sequence 
		       (list
			(make-Postfix trampoline-counter '++)
			(make-Cond (make-Binary trampoline-counter '< trampoline-limit)
				   (make-Call fun-final
					      (list arg-final))
				   (make-New trampoline-class
					     (list fun-final arg-final))))))))))))
	  (if (= 1 (length (Block-stmts trampoline-code)))
	      (car (Block-stmts trampoline-code))
	      trampoline-code))
       (make-Return expr)))
   ;	(make-If (make-Binary trampoline-counter '< trampoline-limit)
   ;		 (make-Return (make-Call fun-var (list arg-var)))
   ;		 (make-Return (make-New trampoline-class
   ;					(list fun-var arg-var))))))))

(define-nmethod (Call-Cps.trans-trampoline!)
   (default-walk! this)
   (make-trampoline-test this))

(define-nmethod (Call-Cont.trans-trampoline!)
   (default-walk! this)
   (make-trampoline-test this))

(define-nmethod (Program.trans-trampoline!)
   (default-walk! this)
   (with-access::Program this (body)
      (set! body (make-trampoline body))) 
   this)

(define dummy-cont  ;;This is a dummy continuation that does
		      ;;nothing.
   (make-Continu '() (make-Block '())))

(define (make-trampoline cps-code)
   (make-Block
    (list (make-Var-Decl-List
	   (list
	    (make-Init trampoline-counter (make-Number 0))
	    (make-Init trampoline-var (make-Call cps-code (list dummy-cont)))))
	  (make-While (make-Binary trampoline-var 'instanceof trampoline-class)
		      (make-Vassig trampoline-var
				   (make-Call (make-Access trampoline-var (make-String "'restart'"))
					      '()))))))

(define-nmethod (Fun.trans-trampoline!)
   (default-walk! this)
   this)

