(module vardecl
   (import js-nodes
	   js-parser
	   js-out
	   js-nodes-extra
	   config
	   verbose
	   walk)
   (export
    (var-decl-normal-form! tree)))

;; The valdecl pass moves all the variable declaration to the begining
;; of a scope. It concerns 1) Variable declaration statement; 2)
;; Variable declaration in For statement 3) Variable declaration in
;; For-In statement 4) Variable declaration in Fun-Binding statement.


(define (var-decl-normal-form! tree)
   (var-decl-nf! tree #f)) 

(define-nmethod (Node.var-decl-nf!) ;; (var-decl-nf! tree #f) only
                                    ;; works on Program node and
                                    ;; Function node, which move all
                                    ;; variable declaration in a scope
                                    ;; to the begining of the scope.
   (error "var-decl-nf!"
	  "var-decl-nf! can only be apply to program node or function node"
	  (class-name (object-class this))))

;(define (var-decl-nf-block! tree) ;;
;   (let ((var-decl-list (cons 0 '())))
;      (var-decl-extract! tree var-decl-list)
;      (verbose "Var-decl-nf-block!: "
;	       (car var-decl-list)
;	       " variable declaration extracted.")
;      (if (> (car var-decl-list) 0)
;	  (make-Var-Decl-Block (Block-stmts tree) (make-Var-Decl-List (cdr var-decl-list)))
;	  (make-Var-Decl-Block (Block-stmts tree) #f))))	  

(define-nmethod (Program.var-decl-nf!) ;;
   (with-access::Program this (body)
      (let ((var-decl-list (cons 0 '())))
	 (var-decl-extract! body var-decl-list)
	 (if (> (car var-decl-list) 0)
	     (make-Var-Program body (make-Var-Decl-List (cdr var-decl-list)))
	     (make-Var-Program body #f)))))

(define-nmethod (Fun.var-decl-nf!) ;;
   (with-access::Fun this (params body)
      (let ((var-decl-list (cons 0 '())))
	 (var-decl-extract! body var-decl-list)
	 (if (> (car var-decl-list) 0)
	     (make-Var-Fun params body (make-Var-Decl-List (cdr var-decl-list)))
	     (make-Var-Fun params body #f)))))  

(define-nmethod (Node.var-decl-extract!) ;; (var-decl-extract! tree
					 ;; var-decl-list) extract all
					 ;; variable declaration in
                                         ;; 'tree', env in
   (default-walk! this)
   this)

(define-nmethod (For.var-decl-extract!)
   (with-access::For this (init test incr body)
      (if init
	  (set! init (var-decl-extract! init env)))
      (if test
	  (set! test (var-decl-extract! test env)))
      (if incr
	  (set! incr (var-decl-extract! incr env)))
      (if body
	  (set! body (var-decl-extract! body env)))
   this))

(define (Init2Vassig tree) ;;Turn a Init node into a Vassig node
   (make-Vassig (Init-lhs tree)
		(Init-rhs tree)))


(define-nmethod (Var-Decl-List.var-decl-extract!) ;; Special case for
						  ;; Var-Decl-List
						  ;; node.
   (default-walk! this)
   (let* ((decl-list (Var-Decl-List-vars this))   ;; Extract a vars
						  ;; list from the
						  ;; node.
	  (decl-list-vars (map (lambda (tree)        ;; Extract all var
						  ;; declarations from
						  ;; the list.
				  (cond ((Init? tree)
					 (Init-lhs tree))
					((Decl? tree)
					 tree)
					(else
					 (error "var-decl-extract!"
						"Unknown type in list of vars"
						(class-name (object-class tree))))))
			       decl-list))
	  (decl-list-inits (filter Init? decl-list)) ;; Filter all
						     ;; inits from the
						     ;; list.
	  (exprs (map Init2Vassig decl-list-inits))) ;; Map to a list
						     ;; of assignment
						     ;; expression.
      (begin                                         ;; Update global
						     ;; env, which is
						     ;; var-decl-list.
	 (set-car! env (+ (car env) (length decl-list-vars)))
	 (set-cdr! env (append (cdr env) decl-list-vars)))
      (if (not (pair? exprs))                        ;; If there is no
						     ;; inits in vars
						     ;; list then make
						     ;; a dummy
						     ;; expression.
	  (set! exprs (cons (make-Undefined 0) exprs)))
      (make-Sequence exprs)))                        ;; Finally return
						     ;; a sequence of
						     ;; expression
(define (Fun-Binding2Vassig tree) ;;Turn a Init treeode into a Vassig node
   (make-Vassig (Fun-Binding-lhs tree)
		(Fun-Binding-rhs tree)))

(define (Decl2Ref tree) ;;Turn a Init node into a Vassig node
   (make-Ref (Decl-id tree)))


(define-nmethod (Fun-Binding.var-decl-extract!) ;; Special case for
						;; Fun-Binding
   (default-walk! this)
   (let ((decl (Fun-Binding-lhs this)))
      (begin                              ;; Update global env, which
					  ;; is var-decl-list.
	 (set-car! env (+ (car env) 1))
	 (set-cdr! env (append (cdr env) (list decl)))))
   (Fun-Binding2Vassig this))

(define-nmethod (For-In.var-decl-extract!)  ;; Special case for For-In
					    ;; node, since we have to
					    ;; keep the var in the
					    ;; position where it was
   (with-access::For-In this (lhs obj body)
      ;(set! lhs (walk! lhs))
      (if (Var-Decl-List? lhs)
	  (let* ((lhs_old lhs)
		 (decl-list (Var-Decl-List-vars lhs))
		 (first-var (car decl-list)))
	     (cond ((Init? first-var)
		    (set! lhs (Decl2Ref (Init-lhs first-var))))
		   ((Decl? first-var)
		    (set! lhs (Decl2Ref first-var)))
		   (else (error "var-decl-extract!"
				"Unknown type in list of vars"
				(class-name (object-class first-var)))))
	     (walk! lhs_old)))
      (set! obj (walk! obj))
      (set! body (walk! body))
      this))
   

(define-nmethod (Return.var-decl-extract!)
   (with-access::Return this (val)
      (if val
	  (set! val (walk0! val env var-decl-extract!)))
      this))

(define-nmethod (Fun.var-decl-extract!) ;; Special case for Fun node
					;; since it create a new scope
   (var-decl-nf! this #f))
   

