(module walk
   (import js-nodes)
   (export
    (generic walk0 n::Node env p::procedure)
    (generic walk1 n::Node env p::procedure arg0)
    (generic walk2 n::Node env p::procedure arg0 arg1)
    (generic walk3 n::Node env p::procedure arg0 arg1 arg2)
    (generic walk4 n::Node env p::procedure arg0 arg1 arg2 arg3)
    (generic walk0! n::Node env p::procedure)
    (generic walk1! n::Node env p::procedure arg0)
    (generic walk2! n::Node env p::procedure arg0 arg1)
    (generic walk3! n::Node env p::procedure arg0 arg1 arg2)
    (generic walk4! n::Node env p::procedure arg0 arg1 arg2 arg3)
    (macro define-nmethod)
    (macro ncall)))

;; (define-nmethod (While.optim! x y) BODY)
;; =>
;; (define-method (optim! this::While env x y)
;;   (define (default-walk! n x y)
;;      (walk2! n env optim! x y))
;;   (define (walk! n x y)
;;      (optim! n env x y))
;;   BODY)
(define-macro (define-nmethod args . body)
   (define (without-type sym)
      (if (not (symbol? sym))
	  (scheme2js-error "bad define-nmethod (expected symbol)" args args)
	  (let* ((str (symbol->string sym))
		 (pos (string-contains str "::")))
	     (if pos
		 (string->symbol (substring str 0 pos))
		 sym))))

   (let* ((Type.name (car args))
	  (method-args (cdr args))
	  (str-type.name (symbol->string Type.name))
	  (str-len (string-length str-type.name))
	  (dot-pos (string-index str-type.name #\.))
	  (dummy (when (not dot-pos)
		    (scheme2js-error "bad define-nmethod name" Type.name Type.name)))
	  (type (string->symbol (substring str-type.name 0 dot-pos)))
	  (name (string->symbol (substring str-type.name
					   (+fx dot-pos 1)
					   str-len)))
	  (bang? (char=? #\! (string-ref str-type.name (- str-len 1))))
	  (nb-method-args (length method-args))
	  (short-walk (if bang? 'walk! 'walk))
	  (long-walk (if bang?
			 (string->symbol (format "walk~a!" nb-method-args))
			 (string->symbol (format "walk~a" nb-method-args))))
	  (define-gen/met (if (eq? type 'Node)
			      'define-generic
			      'define-method))
	  (default-walk (symbol-append 'default- short-walk))
	  (args-without-type (map without-type method-args))
	  (gensymed-method-args (map gensym args-without-type)))
      `(,define-gen/met (,name ,(symbol-append 'this:: type) env ,@method-args)
	  (define (,default-walk node ,@gensymed-method-args)
	     (,long-walk node env ,name ,@gensymed-method-args))
	  (define (,short-walk node ,@gensymed-method-args)
	     (,name node env ,@gensymed-method-args))
	  ,@body)))

(define-macro (ncall method-name node . args)
   `(,method-name ,node env ,@args))
    
(define-generic (walk0 n::Node env p::procedure)
   (error "walk0"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk1 n::Node env p::procedure arg0)
   (error "walk1"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk2 n::Node env p::procedure arg0 arg1)
   (error "walk2"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk3 n::Node env p::procedure arg0 arg1 arg2)
   (error "walk3"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk4 n::Node env p::procedure arg0 arg1 arg2 arg3)
   (error "walk4"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))

(define-generic (walk0! n::Node env p::procedure)
   (error "walk0!"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk1! n::Node env p::procedure arg0)
   (error "walk1!"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk2! n::Node env p::procedure arg0 arg1)
   (error "walk2!"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk3! n::Node env p::procedure arg0 arg1 arg2)
   (error "walk3!"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))
(define-generic (walk4! n::Node env p::procedure arg0 arg1 arg2 arg3)
   (error "walk4!"
	  "Internal Error: forgot Node type"
	  (with-output-to-string (lambda () (write-circle n)))))

(define-macro (gen-walks class . fields)
   (define (field-name f)
      (if (pair? f)
	  (car f)
	  (without-? f)))

   (define (starts-with-? f)
      (char=? #\? (string-ref (symbol->string f) 0)))

   (define (without-? f)
      (if (starts-with-? f)
	  (let ((str (symbol->string f)))
	     (string->symbol (substring str 1 (string-length str))))
	  f))

   (define (visit f nb-args)
      (cond
	 ((pair? f)
	  `(for-each (lambda (f)
			(p f env
			   ,@(map (lambda (i)
				     (string->symbol (format "arg~a" i)))
				  (iota nb-args))))
		     ,(car f)))
	 ((starts-with-? f)
	  `(when ,(without-? f)
	      ,(visit (without-? f) nb-args)))
	 (else
	  `(p ,f env ,@(map (lambda (i)
			       (string->symbol (format "arg~a" i)))
			    (iota nb-args))))))

   (define (visit! f nb-args)
      (cond
	 ((pair? f)
	  `(let loop ((fields ,(car f)))
	      (unless (null? fields)
		 (set-car! fields
			   (p (car fields) env
			      ,@(map (lambda (i)
					(string->symbol (format "arg~a" i)))
				     (iota nb-args))))
		 (loop (cdr fields)))))
	 ((starts-with-? f)
	  `(when ,(without-? f)
	      ,(visit! (without-? f) nb-args)))
	 (else
	  `(set! ,f (p ,f env ,@(map (lambda (i)
					(string->symbol (format "arg~a" i)))
				     (iota nb-args)))))))
   
   (define (gen-method nb-args bang?)
      (let ((val
	     `(define-method (,(if bang?
				   (string->symbol (format "walk~a!" nb-args))
				   (string->symbol (format "walk~a" nb-args)))
			      ,(symbol-append 'n:: class)
			      env
			      p
			      ,@(map (lambda (i)
					(string->symbol (format "arg~a" i)))
				     (iota nb-args)))
		 ,(if (null? fields)
		      #unspecified
		      `(,(symbol-append 'with-access:: class) n ,(map field-name fields)
							      ,@(map (lambda (f)
									((if bang? visit! visit) f nb-args))
								     fields)))
		 n)))
	 ;(tprint val)
	 val))

   `(begin
       ,@(map (lambda (nb) (gen-method nb #f)) (iota 4))
       ,@(map (lambda (nb) (gen-method nb #t)) (iota 4))))

(gen-walks Program body)
(gen-walks Block (stmts))
(gen-walks Ref)
(gen-walks This)
(gen-walks Decl)
(gen-walks Var-Decl-List (vars))
(gen-walks NOP)
(gen-walks If test then else)
(gen-walks For init test incr)
(gen-walks For-In lhs obj body)
(gen-walks While test body)
(gen-walks Do body test)
(gen-walks Continue)
(gen-walks Break)
(gen-walks Return val)
(gen-walks With obj body)
(gen-walks Switch key (cases))
(gen-walks Case expr body)
(gen-walks Default body)
(gen-walks Throw expr)
(gen-walks Try body catch finally)
(gen-walks Catch exception body)
(gen-walks Labeled body)
(gen-walks Fun-Binding lhs rhs)
(gen-walks Assig lhs rhs)
(gen-walks Assig-op lhs rhs)
(gen-walks Vassig lhs rhs)
(gen-walks Accsig lhs rhs)
(gen-walks Vassig-op lhs rhs)
(gen-walks Accsig-op lhs rhs)
(gen-walks Init lhs rhs)
(gen-walks Param)
(gen-walks Named-Fun name fun)
(gen-walks Fun (params) body)
(gen-walks Sequence (exprs))
(gen-walks Cond test then else)
(gen-walks Binary lhs rhs)
(gen-walks Unary expr)
(gen-walks Postfix expr)
(gen-walks New class (args))
(gen-walks Access obj field)
(gen-walks Call fun (args))
(gen-walks Pragma)
(gen-walks Array (els))
(gen-walks Array-Element expr)
(gen-walks Obj-Init (inits))
(gen-walks Property-Init val)
(gen-walks Literal)
(gen-walks String)
(gen-walks Number)
(gen-walks Undefined)
(gen-walks Null)
(gen-walks Bool)
(gen-walks RegExp)
